# coding=utf8
""" Constants

Shared constants
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2023-03-09"

EMPTY_TUUID = '00000000000040008000000000000000'
"""Empty, but valid, Trimmed UUID"""
EMPTY_UUID = '00000000-0000-4000-8000-000000000000'
"""Empty, but valid, Full UUID"""

SECONDS_HOUR = 3600
SECONDS_DAY = 86400
SECONDS_WEEK = 604800
"""Seconds related constants"""